#!/usr/bin/env python

# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import os
import tempfile
import shutil

import xnat
from xnat.exceptions import XNATLoginFailedError

def main():
    parser = argparse.ArgumentParser(description='Copy Xnat projects')
    parser.add_argument('--host', type=str, required=True, help='XNAT url')
    parser.add_argument('--folder', type=str, required=True, help='path to upload dir')

    parser.add_argument('--destination', type=str, required=False, default=None, help='archive/prearchive')
    parser.add_argument('--retries', type=str, required=False, default=None, help='number of retries for upload')
    parser.add_argument('--project', type=str, required=False, default=None, help='project id')
    parser.add_argument('--subject', type=str, required=False, default=None, help='subject label')
    parser.add_argument('--experiment', type=str, required=False, default=None,  help='experiment label')

    parser.add_argument('--import-handler', type=str, )
    parser.add_argument('--quarantine', type=bool, required=False, default=False, help='Quarantine data')
    parser.add_argument('--trigger-pipelines', type=bool, required=None, default=False, help='Trigger pipelines')

    args = parser.parse_args()
    try:
        with xnat.connect(args.host) as xnat_host:
            xnat_host.services.import_dir(args.folder, quarantine=args.quarantine, destination=args.destination,
                                          trigger_pipelines=args.trigger_pipelines, project=args.project, subject=args.subject,
                                          experiment=args.experiment, import_handler=None)
    except XNATLoginFailedError:
        print(f"ERROR Failed to login")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import argparse
import os
import tempfile
import shutil

import six
from six.moves import html_parser

import xnat


class XNATProjectCopier:
    def __init__(self, source_xnat, source_project, dest_xnat, dest_project):
        self.source_xnat = source_xnat
        self.source_project = source_project
        self.dest_xnat = dest_xnat
        self.dest_project = dest_project
        self.temp_dir = tempfile.mkdtemp()
        print('Using tmpdir: {}'.format(self.temp_dir))

    def __del__(self):
        shutil.rmtree(self.temp_dir)
        pass

    def copy_fields(self, source, destination, prefix=''):
        for field_id, value in source.fields.items():
            # Avoid double escaping of html chars
            destination.fields[field_id] = html_parser.HTMLParser().unescape(value)
            print('{prefix}copying field: {}'.format(
                field_id, prefix=prefix
            ))

    def copy_demographics(self, source_subject, dest_subject):
        """
        Copy over all demographics for a subject
        """
        print('  copying demographics')
        demographic_list = [
            'age', 'birth_weight', 'dob', 'education', 'education_desc', 'employment',
            'ethnicity', 'gender', 'gestational_age', 'handedness', 'height', 'post_menstrual_age',
            'race', 'race2', 'race3', 'race4', 'race5', 'race6', 'ses', 'weight', 'yob']

        demographics_data = {}

        for demographic in demographic_list:
            value = source_subject.demographics.data.get(demographic)
            if value is not None:
                demographics_data[demographic] = value

        dest_subject.demographics.mset(demographics_data)

    def copy_resource(self, source_resource, dest_resource, prefix=''):
        # Download resource content, and upload it again
        print('{prefix}copying resource {}'.format(source_resource.label, prefix=prefix))

        source_resource.download_dir(self.temp_dir, verbose=False)
        for path, dirnames, filenames in os.walk(self.temp_dir):
            if source_resource.label in dirnames and os.path.split(path)[-1] == 'resources':
                resource_path = os.path.join(path, source_resource.label, 'files')
                if os.path.exists(resource_path):
                    break
        else:
            raise ValueError('Could not find directory for downloaded resource!')

        # Upload entire resource directory
        dest_resource.upload_dir(resource_path, method='tgz_memory')

    def copy_resources(self, source_object, dest_object, prefix=''):
        for source_resource in source_object.resources.values():
            resource_id = source_resource.label
            if resource_id in dest_object.resources:
                print('{}Skipping resource {}'.format(prefix, resource_id))
                # Already there, do not copy
                continue

            # Create a resources of the same xsitype
            print('{}Copying resource {}'.format(prefix, resource_id))
            dest_class = self.dest_xnat.XNAT_CLASS_LOOKUP[source_resource.__xsi_type__]
            dest_resource = dest_class(
                parent=dest_object,
                label=source_resource.label,
                content=source_resource.content
            )

            # Copy resource file contents
            if len(source_resource.files) > 0:
                self.copy_resource(source_resource, dest_resource, prefix=prefix)

    def copy_experiment(self, source_experiment, dest_experiment):
        self.copy_fields(source_experiment, dest_experiment, prefix='    ')
        self.copy_resources(source_experiment, dest_experiment, prefix='    ')

        for scan in source_experiment.scans.values():
            print('    copying scan {} / {}'.format(scan.id, scan.type))
            if scan.id not in dest_experiment.scans:
                dest_class = self.dest_xnat.XNAT_CLASS_LOOKUP.get(scan.__xsi_type__)
                if dest_class is None:
                    print('     [WARNING] {} class not found on destination server, skipping'.format(source_experiment.__xsi_type__))
                    continue

                print('     creating scan {}'.format(scan.id))
                dest_scan = dest_class(parent=dest_experiment, id=scan.id, type=scan.type)
            else:
                dest_scan = dest_experiment.scans[scan.id]

            self.copy_resources(scan, dest_scan, prefix='     ')

        for assessor_id, source_assessor in source_experiment.assessors.items():
            # Create an assessor of the same xsitype
            if source_assessor.__xsi_type__ != 'xnat:mrAssessorData':
                dest_class = self.dest_xnat.XNAT_CLASS_LOOKUP[source_assessor.__xsi_type__]
            else:
                print("    Avoiding creation of invalid class xnat:mrAssessorData,"
                      " substituting with xnat:qcAssessmentData")
                dest_class = self.dest_xnat.classes.QcAssessmentData
            dest_assessor = dest_class(parent=dest_experiment, label=source_assessor.label)
            print('    copying assessor {}'.format(source_assessor.label))
            self.copy_resources(source_assessor, dest_assessor, prefix='      ')

    def copy_subject(self, source_subject, dest_subject):
        self.copy_demographics(source_subject, dest_subject)
        self.copy_fields(source_subject, dest_subject, prefix='  ')
        self.copy_resources(source_subject, dest_subject, prefix='  ')

        for source_experiment in source_subject.experiments.values():
            print('  copying experiment  {}'.format(source_experiment.label))

            if hasattr(source_experiment, 'scans') and len(source_experiment.scans) > 0:
                print('    copying data')
                temp_file = os.path.join(self.temp_dir, source_experiment.label + '.zip')
                try:
                    source_experiment.download(temp_file, verbose=False)
                    try:
                        dest_experiment = self.dest_xnat.services.import_(
                            temp_file,
                            project=self.dest_project.id,
                            subject=source_experiment.subject.label,
                            experiment=source_experiment.label
                        )
                    except xnat.exceptions.XNATUploadError as exception:
                        print('    [ WARNING] Experiment did not include parsable dicom files, creating empty experiment')
                        if 'not include parseable files' in exception.args[0]:
                            dest_class = self.dest_xnat.XNAT_CLASS_LOOKUP.get(source_experiment.__xsi_type__)
                            if dest_class is None:
                                print('    [WARNING] {} class not found on destination server, skipping'.format(source_experiment.__xsi_type__))
                                continue

                            dest_experiment = dest_class(parent=dest_subject, label=source_experiment.label)
                        else:
                            raise
                finally:
                    try:
                        os.remove(temp_file)
                    except:
                        pass  # Allowed

            else:
                print('    creating empty experiment')
                dest_class = self.dest_xnat.XNAT_CLASS_LOOKUP.get(source_experiment.__xsi_type__)
                if dest_class is None:
                    print('    [WARNING] {} class not found on destination server, skipping'.format(source_experiment.__xsi_type__))
                    continue

                dest_experiment = dest_class(parent=dest_subject, label=source_experiment.label)
            self.copy_experiment(source_experiment, dest_experiment)

    def copy_project(self):
        source_project = self.source_project
        dest_project = self.dest_project

        print('Copying fields')
        self.copy_fields(source_project, dest_project)
        print('Copying resources')
        self.copy_resources(source_project, dest_project, prefix='  ')

        for subject_id, source_subject in self.source_project.subjects.items():
            print('copying subject {}'.format(source_subject.label))
            dest_subject = self.dest_xnat.classes.SubjectData(parent=self.dest_project, label=source_subject.label)
            self.copy_subject(source_subject, dest_subject)

    def start(self):
        self.copy_project()


def main():
    parser = argparse.ArgumentParser(description='Copy Xnat projects')
    parser.add_argument('--source-host', type=six.text_type, required=True, help='source XNAT url')
    parser.add_argument('--source-project', type=six.text_type, required=True, help='source XNAT project')
    parser.add_argument('--dest-host', type=six.text_type, required=True, help='destination XNAT url')
    parser.add_argument('--dest-project', type=six.text_type, required=True, help='destination XNAT project')
    args = parser.parse_args()

    with xnat.connect(args.source_host) as source_xnat, xnat.connect(args.dest_host) as dest_xnat:
        # Find projects
        try:
            source_project = source_xnat.projects[args.source_project]
            dest_project = dest_xnat.projects[args.dest_project]
        except KeyError as error:
            print(error.message)
        else:
            # Create and start copier
            copier = XNATProjectCopier(source_xnat, source_project, dest_xnat, dest_project)
            copier.start()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
from time import sleep
import threading
from pathlib import Path
import abc

import json
import xnat
from xnat.exceptions import XNATLoginFailedError


class FileObject:
    # def __init__(self, project, subject, experiment, scan, resource, filename, count, location):
    def __init__(self, location: str, parent_info: dict, filename: str = None, count: int = 1):
        self.count = count
        self._location = location
        self._data = {
            'project': parent_info.get('project'),
            'subject': parent_info.get('subject'),
            'experiment': parent_info.get('experiment'),
            'scan': parent_info.get('scan'),
            'resource': parent_info.get('resource'),
            'assessor': parent_info.get('assessor'),
            'filename': filename,
        }

    def __hash__(self):
        return hash((
            self._data['project'],
            self._data['subject'],
            self._data['experiment'],
            self._data['scan'],
            self._data['resource'],
            self._data['assessor'],
            self._data['filename'],
        ))

    def __eq__(self, other) -> bool:
        return self._data == other._data

    def __str__(self) -> str:
        value_list = [
            self._location,
            str(self._data['project']),
            str(self._data['subject']),
            str(self._data['experiment']),
            str(self._data['scan']),
            str(self._data['resource']),
            str(self._data['assessor']),
            str(self._data['filename']),
            str(self.count),
        ]
        return '\t'.join(value_list)


class BaseScraper:
    def __init__(self, name: str):
        self._name = name
        self._work_list = []
        self.files = set()
        self._progress = 0

    def start(self):
        self._generate_worklist()
        self._check_running = True
        self._check_thread = threading.Thread(target=self._scrape, name=self._name)
        self._check_thread.daemon = True
        self._check_thread.start()

    def is_running(self) -> bool:
        return self._check_thread.is_alive()

    def progress(self) -> float:
        return (self._progress/len(self._work_list))

    @abc.abstractmethod
    def _generate_worklist(self):
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def _scrape(self):
        raise NotImplementedError("Please Implement this method")


class XNATFileScraper(BaseScraper):
    def __init__(self, host: str):
        super().__init__('XNATFileScraper')
        self.host = host
        self.xnat_host = xnat.connect(host)
        self.missing_data = []

    def _generate_worklist(self):
        self._xnat_progress = 0
        self._work_list = self.xnat_host.projects
        print(f"XNAT contains {len(self._work_list)} projects: {' '.join(self._work_list)}")

    def _scrape(self):
        print(f"Start XNAT check on {self.xnat_host.uri}")
        for project in self._work_list:
            xnat_project = self.xnat_host.projects[project]
            for subject_id, xnat_subject in xnat_project.subjects.items():
                parent_info = {
                    'project': project,
                }
                self._check_xnat_subject(dict(parent_info), xnat_subject)

            self._progress += 1

    def _check_xnat_resource(self, parent_info: dict, xnat_resource):
        parent_info['resource'] = xnat_resource.label
        if xnat_resource.file_count != len(xnat_resource.files):
            xnat_file_info = FileObject('xnat', parent_info, count=xnat_resource.file_count - len(xnat_resource.files))
            self.files.add(xnat_file_info)
        for file_id, xnat_file in xnat_resource.files.items():
            xnat_file_info = FileObject('xnat', parent_info, filename=xnat_file.path)
            self.files.add(xnat_file_info)

    def _check_xnat_experiment(self, parent_info: dict, xnat_experiment):
        parent_info['experiment'] = xnat_experiment.label
        for resource_id, xnat_resource in xnat_experiment.resources.items():
            result = self._check_xnat_resource(dict(parent_info), xnat_resource)

        for assessor_id, xnat_assessor in xnat_experiment.assessors.items():
            result = self._check_xnat_assessor(dict(parent_info), xnat_assessor)

        for scan_id, xnat_scan in xnat_experiment.scans.items():
            self._check_xnat_scan(dict(parent_info), xnat_scan)

    def _check_xnat_assessor(self, parent_info: dict, xnat_assessor):
        parent_info['assessor'] = xnat_assessor.label
        for resource_id, xnat_resource in xnat_assessor.resources.items():
            result = self._check_xnat_resource(dict(parent_info), xnat_resource) 

    def _check_xnat_scan(self, parent_info: dict, xnat_scan):
        parent_info['scan'] = xnat_scan.id
        for resource_id, xnat_resource in xnat_scan.resources.items():
            result = self._check_xnat_resource(dict(parent_info), xnat_resource)

    def _check_xnat_subject(self, parent_info: dict, xnat_subject):
        for experiment_id, xnat_experiment in xnat_subject.experiments.items():
            result = self._check_xnat_experiment(dict(parent_info), xnat_experiment)
        # Pass subject only to resource data not needed for FileSystem comparison.
        # No subject info available for experiments on Filesystem
        parent_info['subject'] = xnat_subject.label
        for resource_id, xnat_resource in xnat_subject.resources.items():
            result = self._check_xnat_resource(dict(parent_info), xnat_resource)


class FileSystemFileScraper(BaseScraper):
    def __init__(self, base_dir):
        super().__init__('FileSystemFileScraper')
        self.xnat_base_dir = Path(base_dir)

    def _generate_worklist(self):
        self._fs_progress = 0
        self._work_list = [x for x in self.xnat_base_dir.iterdir() if x.is_dir()]
        projects = [str(x) for x in self._work_list]
        print(f"FS contains {len(self._work_list)} projects: {' '.join(projects)}")

    def _scrape(self):
        print(f"Start filesystem check on {self.xnat_base_dir}")
        for project in self._work_list:
            # Check experiment data
            experiments = [x for x in (project / 'arc001').iterdir() if x.is_dir()]
            for experiment in experiments:
                parent_info = {
                    'project': project.name
                }
                self._fs_experiment_check(dict(parent_info), experiment)

            # Check subject specific data(resources)
            if (project / 'subjects').exists():
                subjects = [x for x in (project / 'subjects').iterdir() if x.is_dir()]
                for subject in subjects:
                    parent_info = {
                        'project': project.name
                    }
                    self._fs_subject_check(dict(parent_info), subject)

            self._progress += 1

    def _fs_resource_check(self, parent_info: dict, resource):
        parent_info['resource'] = resource.name
        files = [x for x in resource.iterdir() if x.is_file()]
        for file in files:
            if file.name.endswith('catalog.xml'):
                continue
            xnat_file_info = FileObject('filesystem', parent_info, filename=str(file.name))
            self.files.add(xnat_file_info)

    def _fs_subject_check(self, parent_info: dict, subject):
        parent_info['subject'] = subject.name
        resources = [x for x in subject.iterdir() if x.is_dir()]
        for resource in resources:
            self._fs_resource_check(dict(parent_info), resource)

    def _fs_experiment_check(self, parent_info: dict, experiment):
        parent_info['experiment'] = experiment.name

        resources_dir = experiment / 'RESOURCES'
        if resources_dir.is_dir():
            resources = [x for x in resources_dir.iterdir() if x.is_dir()]
            for resource in resources:
                self._fs_resource_check(dict(parent_info), resource)

        assessors_dir = experiment / 'ASSESSORS'
        if assessors_dir.is_dir():
            assessors = [x for x in assessors_dir.iterdir() if x.is_dir()]
            for assessor in assessors:
                self._fs_assessor_check(dict(parent_info), assessor)

        scans_dir = experiment / 'SCANS'
        if scans_dir.is_dir():
            scans = [x for x in scans_dir.iterdir() if x.is_dir()]
            for scan in scans:
                self._fs_scan_check(dict(parent_info), scan)

    def _fs_assessor_check(self, parent_info: dict, assessor):
        parent_info['assessor'] = assessor.name
        resources = [x for x in assessor.iterdir() if x.is_dir()]
        for resource in resources:
            self._fs_resource_check(dict(parent_info), resource)

    def _fs_scan_check(self, parent_info: dict, scan):
        parent_info['scan'] = scan.name
        resources = [x for x in scan.iterdir() if x.is_dir()]
        for resource in resources:
            self._fs_resource_check(dict(parent_info), resource)


class XNATIntegrityCheck:
    def __init__(self, host, xnat_base_dir):
        self._xnat_file_scraper = XNATFileScraper(host)
        self._filesystem_scraper = FileSystemFileScraper(xnat_base_dir)

    def start(self):
        self._xnat_file_scraper.start()
        self._filesystem_scraper.start()

    def is_running(self):
        return self._xnat_file_scraper.is_running() or self._filesystem_scraper.is_running()

    def progress(self):
        return (self._xnat_file_scraper.progress(), self._filesystem_scraper.progress())

    def write_report(self, filename):
        files_on_xnat = sum(file_object.count for file_object in self._xnat_file_scraper.files)
        files_on_fs = sum(file_object.count for file_object in self._filesystem_scraper.files)
        print(f'Files scanned on XNAT: {files_on_xnat}')
        print(f'Files scanned on FS: {files_on_fs}')

        missing_on_xnat = self._filesystem_scraper.files.difference(self._xnat_file_scraper.files)
        missing_on_xnat_count = sum(missing.count for missing in missing_on_xnat)
        print(f'Files missing on XNAT: {missing_on_xnat_count}')

        missing_on_fs = self._xnat_file_scraper.files.difference(self._filesystem_scraper.files)
        missing_on_fs_count = sum(missing.count for missing in missing_on_fs)
        print(f'Files missing on FS: {missing_on_fs_count}')

        with open(filename, 'w') as fo:
            header = 'location\tproject\tsubject\texperiment\tscan\tresource\tassessor\tfilename\tcount'
            fo.write(f"{header}\n")
            for file in missing_on_xnat:
                fo.write(f"{file}\n")
            # fo.write(f'Missing on Filesystem:\n')
            for file in missing_on_fs:
                fo.write(f"{file}\n")


def main():
    parser = argparse.ArgumentParser(description='Xnat data integrity check')
    parser.add_argument('--host', type=str, required=True, help='XNAT url')
    parser.add_argument('--xnat-home-dir', type=str, required=True, help='path to xnat home dir')
    parser.add_argument('--report', type=str, required=True, help='path to xnat home dir')
    args = parser.parse_args()

    xnat_integrity_checker = XNATIntegrityCheck(args.host, args.xnat_home_dir)
    xnat_integrity_checker.start()
    print('progress\t FS\tXNAT')
    while xnat_integrity_checker.is_running():
        xnat_progress, fs_progress = xnat_integrity_checker.progress()
        print(f'\t\t{fs_progress*100}\t{xnat_progress*100}', end='\r')
        sleep(1)
    fs_progress, xnat_progress = xnat_integrity_checker.progress()
    print(f'\t\t{fs_progress*100}\t{xnat_progress*100}')
    print("--- REPORT ---")
    xnat_integrity_checker.write_report(args.report)
    print('--- DONE ---')


if __name__ == "__main__":
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The specific exceptions that are thrown by XNATpy
"""

from __future__ import absolute_import
import requests


class XNATError(Exception):
    """
    General base class for all XNAT related errors
    """


class XNATValueError(XNATError, ValueError):
    """
    XNATpy specific value error
    """


class XNATResponseError(XNATValueError):
    """
    XNATpy error when the response value is not correct
    """


class XNATConnectionError(XNATValueError):
    """
    XNATpy error when the connection cannot be established
    """


class XNATIOError(XNATError, IOError):
    """
    XNATpy error for when there are IO problems
    """


class XNATUploadError(XNATIOError):
    """
    XNATpy error for when there is a problem uploading
    """


class XNATSSLError(XNATError, requests.exceptions.SSLError):
    """
    XNATpy error for when there is an SSL problem
    """


# Inherit from ValueError for backwards compatibility (they used to be value errors)
class XNATAuthError(XNATError, ValueError):
    """
    XNATpy error for when there is a problem with logging in or authentication
    """


class XNATLoginFailedError(XNATAuthError):
    """
    Failed to login, this usually means the credentials are incorrect.
    """


class XNATExpiredCredentialsError(XNATAuthError):
    """
    The users credentials are expired and should be updated in the web interface of XNAT
    """


class XNATNotConnectedError(XNATError):
    """
    The connection is not open, either it was never established or it was disconnected.
    """


class XNATObjectDestroyedError(XNATError):
    """
    The object manipulated is deleted and therefore destroyed. It should not be used anymore!
    """


# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
from abc import ABCMeta, abstractproperty
from collections import namedtuple
from six.moves.collections_abc import MutableMapping, MutableSequence, Mapping, Sequence
import fnmatch
import keyword
import re
from functools import update_wrapper

from . import exceptions
from .datatypes import convert_from, convert_to
from .constants import TYPE_HINTS
from .utils import mixedproperty, pythonize_attribute_name
import six


def caching(func):
    """
    This decorator caches the value in self._cache to avoid data to be
    retrieved multiple times. This works for properties or functions without
    arguments.
    """
    name = func.__name__

    def wrapper(self):
        # We use self._cache here, in the decorator _cache will be a member of
        #  the objects, so nothing to worry about
        # pylint: disable=protected-access
        if not self.caching or name not in self._cache:
            # Compute the value if not cached
            self._cache[name] = func(self)

        return self._cache[name]

    update_wrapper(wrapper, func)
    return wrapper


class VariableMap(MutableMapping):
    def __init__(self, parent, field):
        self._cache = {}
        self.caching = True
        self.parent = parent
        self._field = field

    def __repr__(self):
        return "<VariableMap {}>".format(dict(self))

    @property
    @caching
    def data(self):
        try:
            variables = next(x for x in self.parent.fulldata['children'] if x['field'] == self.field)
            variables_map = {x['data_fields']['name']: x['data_fields']['field'] for x in variables['items'] if 'field' in x['data_fields']}
        except StopIteration:
            variables_map = {}

        return variables_map

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        query = {'xsiType': self.parent.__xsi_type__,
                 '{parent_type_}/{field}[@xsi_type={type}]/{key}'.format(parent_type_=self.parent.__xsi_type__,
                                                                         field=self.field,
                                                                         type=self.parent.__xsi_type__,
                                                                         key=key): value}
        self.xnat.put(self.parent.fulluri, query=query)

        # Remove cache and make sure the reload the data
        if 'data' in self._cache:
            self.clearcache()

    def __delitem__(self, key):
        self.parent.logger.warning('Deleting of variables is currently not supported!')

    def __iter__(self):
        for key in self.data.keys():
            yield key

    def __len__(self):
        return len(self.data)

    @property
    def field(self):
        return self._field

    @property
    def xnat(self):
        return self.parent.xnat_session

    def clearcache(self):
        self._cache.clear()
        self.parent.clearcache()


class CustomVariableMap(VariableMap):
    def __setitem__(self, key, value):
        query = {'xsiType': self.parent.__xsi_type__,
                 '{type_}/fields/field[name={key}]/field'.format(type_=self.parent.__xsi_type__,
                                                                 key=key): value}
        self.xnat.put(self.parent.fulluri, query=query)

        # Remove cache and make sure the reload the data
        if 'data' in self._cache:
            self.clearcache()


@six.python_2_unicode_compatible
class XNATBaseObject(six.with_metaclass(ABCMeta, object)):
    SECONDARY_LOOKUP_FIELD = None
    _DISPLAY_IDENTIFIER = None
    _HAS_FIELDS = False
    _CONTAINED_IN = None
    _XSI_TYPE = 'xnat:baseObject'

    def __init__(self, uri=None, xnat_session=None, id_=None, datafields=None, parent=None, fieldname=None, overwrites=None, **kwargs):
        if (uri is None or xnat_session is None) and parent is None:
            raise exceptions.XNATValueError('Either the uri and xnat session have to be given, or the parent object')

        # Set the xnat session
        self._cache = {}
        self._caching = None

        # This is the object creation branch
        if uri is None and parent is not None:
            # This is the creation of a new object in the XNAT server
            self._xnat_session = parent.xnat_session
            if isinstance(parent, XNATListing):
                pass
            elif self._CONTAINED_IN is not None:
                parent = getattr(parent, self._CONTAINED_IN)
            else:
                self.logger.debug('parent {}, self._CONTAINED_IN: {}'.format(parent, self._CONTAINED_IN))
                raise exceptions.XNATValueError('Cannot determine PUT url!')

            # Check what argument to use to build the URL
            if self._DISPLAY_IDENTIFIER is not None:
                url_part_argument = pythonize_attribute_name(self._DISPLAY_IDENTIFIER)
            elif self.SECONDARY_LOOKUP_FIELD is not None:
                url_part_argument = self.SECONDARY_LOOKUP_FIELD
            else:
                raise exceptions.XNATValueError('Cannot figure out correct object creation url for <{}>, '
                                                'creation currently not supported!'.format(type(self).__name__))

            # Get extra required url part
            url_part = six.text_type(kwargs.get(url_part_argument))

            if url_part is not None:
                uri = '{}/{}'.format(parent.uri, url_part)
                self.logger.debug('PUT URI: {}'.format(uri))
                query = {
                    'xsiType': self.__xsi_type__,
                    'req_format': 'qs',
                }

                # Add all kwargs to query with correct xpath to set the fields
                for name, value in kwargs.items():
                    xpath = '{}/{}'.format(self.xpath, name)
                    query[xpath] = value

                self.logger.debug('query: {}'.format(query))
                result = self.xnat_session.put(uri, query=query)
                self.logger.debug('PUT RESULT: [{}] {}'.format(result.status_code, result.text))
                result_text = result.text.strip()

                # This should be the ID of the newly created object, which is safer than
                # labels that can contain weird characters and break stuff
                if result_text:
                    uri = '{}/{}'.format(parent.uri, result_text)
                    self.logger.debug('UPDATED URI BASED ON RESPONSE: {}'.format(uri))
            else:
                raise exceptions.XNATValueError('The {} for a {} need to be specified on creation'.format(
                    url_part_argument,
                    self.__xsi_type__
                ))

            # Clear parent cache
            parent.clearcache()

            # Parent is no longer needed after creation
            self._uri = uri
            self._parent = None

            # Add url part to overwrites (it should be safe) but rest should be retrieved from server to be sure
            # the creation went correctly
            self._overwrites = overwrites or {}
            self._overwrites[url_part_argument] = url_part
        else:
            # This is the creation of a Python proxy for an existing XNAT object
            self._uri = uri
            self._parent = parent

            # Cache the kwargs in the object already
            self._overwrites = overwrites or {}
            self._overwrites.update(kwargs)

        self._xnat_session = xnat_session
        self._fieldname = fieldname

        if self._HAS_FIELDS:
            self._fields = CustomVariableMap(self, field='fields/field')
        else:
            self._fields = None

        if id_ is not None:
            self._cache['id'] = id_

        if datafields is not None:
            self._cache['data'] = datafields

    def __str__(self):
        if self.SECONDARY_LOOKUP_FIELD is None:
            return six.text_type('<{} {}>').format(self.__class__.__name__, self.id)
        else:
            return six.text_type('<{} {} ({})>').format(self.__class__.__name__,
                                                        getattr(self, self.SECONDARY_LOOKUP_FIELD),
                                                        self.id)

    def __repr__(self):
        return str(self)

    @abstractproperty
    def xpath(self):
        """
        The xpath of the object as seen from the root of the data. Used for
        setting fields in the object.
        """

    @property
    def parent(self):
        return self._parent

    @property
    def logger(self):
        return self.xnat_session.logger

    @property
    def fieldname(self):
        return self._fieldname

    def get(self, name, type_=None):
        try:
            value = self._overwrites[name]
        except KeyError:
            value = self.data.get(name)

        if type_ is not None and value is not None:
            if isinstance(type_, six.string_types):
                value = convert_to(value, type_)
            else:
                value = type_(value)
        return value

    def get_object(self, fieldname, type_=None):
        try:
            data = next(x for x in self.fulldata.get('children', []) if x['field'] == fieldname)['items']
            data = next(x for x in data if not x['meta']['isHistory'])  # Filter out the non-history item
            type_ = data['meta']['xsi:type']
        except StopIteration:
            if type_ is None:
                type_ = TYPE_HINTS.get(fieldname, None)

        if type_ is None:
            raise exceptions.XNATValueError('Cannot determine type of field {}!'.format(fieldname))

        cls = self.xnat_session.XNAT_CLASS_LOOKUP[type_]

        if not issubclass(cls, (XNATSubObject, XNATNestedObject)):
            raise ValueError('{} is not a subobject type!'.format(cls))

        return self.xnat_session.create_object(self.uri, type_=type_, parent=self, fieldname=fieldname)

    @property
    def fulluri(self):
        return self.uri

    def external_uri(self, query=None, scheme=None):
        """
        Return the external url for this object, not just a REST path

        :param query: extra query string parameters
        :param scheme: scheme to use (when not using original url scheme)
        :return: external url for this object
        """
        return self.xnat_session.url_for(self, query=query, scheme=scheme)

    def mset(self, values=None, timeout=None, **kwargs):
        if not isinstance(values, dict):
            values = kwargs

        if self.parent is not None:
            xsi_type = self.parent.__xsi_type__
        else:
            xsi_type = self.__xsi_type__

        # Add xpaths to query
        query = {'xsiType': xsi_type}
        for name, value in values.items():
            xpath = '{}/{}'.format(self.xpath, name)
            query[xpath] = value

        self.xnat_session.put(self.fulluri, query=query, timeout=timeout)
        self.clearcache()
        if hasattr(self.parent, 'clearcache'):
            self.parent.clearcache()

    def set(self, name, value, type_=None, timeout=None):
        """
        Set a field in the current object

        :param str name: name of the field
        :param value:  value to set
        :param type_: type of the field
        """
        if type_ is not None:
            if isinstance(type_, six.string_types):
                # Make sure we have a valid string here that is properly casted
                value = convert_from(value, type_)
            else:
                value = type_(value)

        self.mset({name: value}, timeout=timeout)

    def del_(self, name):
        self.mset({name: 'NULL'})

    @mixedproperty
    def __xsi_type__(self):
        return self._XSI_TYPE

    @property
    @caching
    def id(self):
        object_id = self.data.get('ID', None)
        if object_id is not None:
            return object_id
        elif self.parent is not None:
            return '{}/{}'.format(self.parent.id, self.fieldname)
        elif hasattr(self, '_DISPLAY_IDENTIFIER') and self._DISPLAY_IDENTIFIER is not None:
            return getattr(self, self._DISPLAY_IDENTIFIER)
        else:
            return '#NOID#'

    @abstractproperty
    def data(self):
        """
        The data of the current object (data fields only)
        """

    @abstractproperty
    def fulldata(self):
        """
        The full data of the current object (incl children, meta etc)
        """

    @property
    def xnat_session(self):
        if self._uri is None:
            raise exceptions.XNATObjectDestroyedError('This object is delete and cannot be used anymore!')

        return self._xnat_session

    @property
    def uri(self):
        if self._uri is None:
            raise exceptions.XNATObjectDestroyedError('This object is delete and cannot be used anymore!')

        return self._uri

    def clearcache(self):
        self._overwrites.clear()
        self._cache.clear()

    # This needs to be at the end of the class because it shadows the caching
    # decorator for the remainder of the scope.
    @property
    def caching(self):
        if self._caching is not None:
            return self._caching
        else:
            return self.xnat_session.caching

    @caching.setter
    def caching(self, value):
        self._caching = value

    @caching.deleter
    def caching(self):
        self._caching = None

    def delete(self, remove_files=True):
        """
        Remove the item from XNATSession
        """
        query = {}

        if remove_files:
            query['removeFiles'] = 'true'

        # Try to remove object thoroughly
        self.xnat_session.delete(self.fulluri, query=query)
        self.xnat_session.remove_object(self)
        self._uri = None
        self._xnat_session = None

        # Make sure there is no cache, this will cause XNATObjectDestroyedError on subsequent use
        # of this object, indicating that is has been in fact removed
        self.clearcache()


class XNATObject(XNATBaseObject):
    @property
    @caching
    def fulldata(self):
        return next(x for x in self.xnat_session.get_json(self.uri)['items'] if not x['meta']['isHistory'])

    @property
    def data(self):
        return self.fulldata['data_fields']

    @property
    def xpath(self):
        return '{}'.format(self.__xsi_type__)


class XNATNestedObject(XNATBaseObject):
    @property
    def fulldata(self):
        try:
            if isinstance(self.parent.fulldata, dict):
                data = next(x for x in self.parent.fulldata['children'] if x['field'] == self.fieldname)['items']
                data = next(x for x in data if not x['meta']['isHistory'])
            elif isinstance(self.parent.fulldata, list):
                if self.parent.secondary_lookup_field is not None:
                    data = next(x for x in self.parent.fulldata if x['data_fields'][self.parent.secondary_lookup_field] == self.fieldname)
                else:
                    # Just simply select the index
                    data = self.parent.fulldata[self.fieldname]
            else:
                raise ValueError("Found unexpected data in parent! ({})".format(self.parent.fulldata))

        except StopIteration:
            data = {'data_fields': {}}

        return data

    @property
    def data(self):
        return self.fulldata['data_fields']

    @property
    def uri(self):
        return self.parent.uri

    @property
    def xpath(self):
        if isinstance(self.parent, XNATBaseObject):
            return '{}/{}[@xsi:type={}]'.format(self.parent.xpath,
                                                self.fieldname,
                                                self.__xsi_type__)
        else:
            return '{}[{}={}]'.format(self.parent.xpath,
                                      self.parent.secondary_lookup_field,
                                      self.fieldname)

    def clearcache(self):
        super(XNATNestedObject, self).clearcache()
        self.parent.clearcache()


class XNATSubObject(XNATBaseObject):
    _PARENT_CLASS = None

    @property
    def uri(self):
        return self.parent.fulluri

    @property
    def __xsi_type__(self):
        return self.parent.__xsi_type__

    @property
    def xpath(self):
        if isinstance(self.parent, XNATBaseObject):
            # XPath is this plus fieldname
            return '{}/{}'.format(self.parent.xpath, self.fieldname)
        elif isinstance(self.parent, XNATBaseListing):
            # XPath is an index in a list
            if isinstance(self.fieldname, int):
                return '{}[{}]'.format(self.parent.xpath,
                                       self.fieldname)
            else:
                return '{}[{}={}]'.format(self.parent.xpath,
                                          self.parent.secondary_lookup_field,
                                          self.fieldname)
        else:
            raise TypeError('Type of parent is invalid! (Found {})'.format(type(self.parent).__name__))

    @property
    def fulldata(self):
        prefix = '{}/'.format(self.fieldname)

        result = self.parent.fulldata

        if isinstance(result, dict):
            result = {k[len(prefix):]: v for k, v in result['data_fields'].items() if k.startswith(prefix)}
            result = {'data_fields': result}
        elif isinstance(result, list):
            try:
                if self.parent.secondary_lookup_field is not None:
                    result = next(x for x in result if x['data_fields'][self.parent.secondary_lookup_field] == self.fieldname)
                else:
                    result = result[self.fieldname]
            except (IndexError, KeyError):
                return {'data_fields': {}}
        else:
            raise ValueError("Found unexpected data in parent! ({})".format(result))

        return result

    @property
    def data(self):
        return self.fulldata['data_fields']

    def clearcache(self):
        super(XNATSubObject, self).clearcache()
        self.parent.clearcache()


@six.python_2_unicode_compatible
class XNATBaseListing(Mapping, Sequence):
    __ALL_LISTINGS__ = []

    def __init__(self, parent, field_name, secondary_lookup_field=None, xsi_type=None, **kwargs):
        # Cache fields
        self._cache = {}
        self._caching = None

        # Save the parent and field name
        self.parent = parent
        self.field_name = field_name

        # Copy parent xnat session for future use
        self._xnat_session = parent.xnat_session

        # Get the lookup field before type hints, they can ruin it for abstract types
        if secondary_lookup_field is None:
            if xsi_type is not None:
                secondary_lookup_field = self.xnat_session.XNAT_CLASS_LOOKUP.get(xsi_type).SECONDARY_LOOKUP_FIELD

        # Make it possible to override the xsi_type for the contents
        if self.field_name not in TYPE_HINTS:
            self._xsi_type = xsi_type
        else:
            self._xsi_type = TYPE_HINTS[field_name]

        # If Needed, try again
        if secondary_lookup_field is None:
            secondary_lookup_field = self.xnat_session.XNAT_CLASS_LOOKUP.get(self._xsi_type).SECONDARY_LOOKUP_FIELD

        self.secondary_lookup_field = secondary_lookup_field

        # Register listing
        self.__ALL_LISTINGS__.append(self)

    def sanitize_name(self, name):
        name = re.sub('[^0-9a-zA-Z]+', '_', name)

        # Change CamelCaseString to camel_case_string
        # Note that addID would become add_id
        name = re.sub("[A-Z]+", lambda x: '_' + x.group(0).lower(), name)
        if name[0] == '_':
            name = name[1:]

        # Avoid multiple underscores (replace them by single underscore)
        name = re.sub("__+", '_', name)

        # Avoid overwriting keywords TODO: Do we want this, as a property it is not a huge problem?
        if keyword.iskeyword(name):
            name += '_'

        return name

    @property
    def xnat_session(self):
        return self._xnat_session

    @abstractproperty
    def data_maps(self):
        """
        The generator function (should be cached) of all the data access
        properties. They are all generated from the same data, so their
        caching is shared.
        """

    @property
    def data(self):
        """
        The data mapping using the primary key
        """
        return self.data_maps[0]

    @property
    def key_map(self):
        """
        The data mapping using the secondary key
        """
        return self.data_maps[1]

    @property
    def non_unique_keys(self):
        """
        Set of non_unique keys
        """
        return self.data_maps[2]

    @property
    def listing(self):
        """
        The listing view of the data
        """
        return self.data_maps[3]

    @abstractproperty
    def xnat_session(self):
        pass

    def __str__(self):
        if self.secondary_lookup_field is not None:
            content = ', '.join('({}, {}): {}'.format(k, getattr(v, self.sanitize_name(self.secondary_lookup_field)), v) for k, v in self.items())
            content = '{{{}}}'.format(content)
        else:
            content = ', '.join(str(v) for v in self.values())
            content = '[{}]'.format(content)
        return '<{} {}>'.format(type(self).__name__, content)

    def __repr__(self):
        return str(self)

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return self.listing[item]

        try:
            return self.data[item]
        except KeyError:
            try:
                if item in self.non_unique_keys:
                    raise KeyError('There are multiple items with that key in'
                                   ' this collection! To avoid problem you need'
                                   ' to use the ID.')
                return self.key_map[item]
            except KeyError:
                raise KeyError('Could not find ID/label {} in collection!'.format(item))

    def __iter__(self):
        # Avoid re-requesting the data for every item
        data = self.data
        for index, item in enumerate(self.listing):
            if hasattr(item, 'id') and item.id in data:
                yield item.id
            elif self.secondary_lookup_field is not None and hasattr(item, self.secondary_lookup_field):
                yield getattr(item, self.secondary_lookup_field)
            else:
                yield index

    def __len__(self):
        return len(self.listing)

    @property
    def uri(self):
        return self._uri

    @property
    def logger(self):
        return self.parent.logger

    @property
    def xnat_session(self):
        return self._xnat_session

    def clearcache(self):
        self._cache.clear()

    # This needs to be at the end of the class because it shadows the caching
    # decorator for the remainder of the scope.
    @property
    def caching(self):
        if self._caching is not None:
            return self._caching
        else:
            return self.xnat_session.caching

    @caching.setter
    def caching(self, value):
        self._caching = value

    @caching.deleter
    def caching(self):
        self._caching = None

    # These two methods allow for wiping the cache of all listings that had obj in them
    # this is required for object deletions so that the objects are no longer presented
    # in the listings
    @classmethod
    def delete_item_from_listings(cls, obj):
        for listing in cls.__ALL_LISTINGS__:
            listing.delete_item_from_cache(obj)

    def delete_item_from_cache(self, obj):
        data_maps = self._cache.get('data_maps', None)

        if data_maps is None:
            return

        if obj in data_maps[0].values() or obj in data_maps[1].values() or obj in data_maps[3]:
            self.clearcache()


class XNATListing(XNATBaseListing):
    def __init__(self, uri, filter=None, **kwargs):
        # Important for communication, needed before superclass is called
        self._uri = uri

        super(XNATListing, self).__init__(**kwargs)

        # Manager the filters
        self._used_filters = filter or {}

    @property
    @caching
    def data_maps(self):
        columns = 'ID,URI'
        if self.secondary_lookup_field is not None:
            columns = '{},{}'.format(columns, self.secondary_lookup_field)
        if self._xsi_type is None:
            columns += ',xsiType'

        query = dict(self.used_filters)
        query['columns'] = columns
        result = self.xnat_session.get_json(self.uri, query=query)

        try:
            result = result['ResultSet']['Result']
        except KeyError:
            raise exceptions.XNATValueError('Query GET from {} returned invalid data: {}'.format(self.uri, result))

        parent_id = None
        for entry in result:
            if 'URI' not in entry and 'ID' not in entry:
                # HACK: This is a Resource, that misses the URI and ID field (let's fix that)
                entry['ID'] = entry['xnat_abstractresource_id']
                entry['URI'] = '{}/{}'.format(self.uri, entry['xnat_abstractresource_id'])
            elif 'ID' not in entry:
                # HACK: This is a File and it misses an ID field and has Name (let's fix that)
                entry['ID'] = entry['Name']
                entry['fieldname'] = type(self.parent).__name__
                if entry['URI'].startswith(self.parent.uri):
                    entry['path'] = entry['URI'].replace(self.parent.uri, '', 1)
                else:
                    if parent_id is None:
                        parent_id = self.parent.id
                    entry['path'] = re.sub(r'^.*/resources/{}/files/'.format(parent_id), '', entry['URI'], 1)
            else:
                entry['URI'] = '{}/{}'.format(self.uri, entry['ID'])

        # Post filter result if server side query did not work
        if self.used_filters:
            result = [x for x in result if all(fnmatch.fnmatch(x[k], v) for k, v in self.used_filters.items() if k in x)]

        # Create object dictionaries
        id_map = {}
        key_map = {}
        listing = []
        non_unique = {None}
        for x in result:
            xsi_type = x.get('xsiType', x.get('element_name', self._xsi_type)).strip()
            if x['ID'].strip() == "" or xsi_type == "":
                self.logger.warning("Found empty object {}, skipping!".format(x.get('URI')))
                continue

            # HACK: xsi_type of resources is called element_name... yay!
            if self.secondary_lookup_field is not None:
                secondary_lookup_value = x.get(self.secondary_lookup_field)
                new_object = self.xnat_session.create_object(x['URI'],
                                                             type_=xsi_type,
                                                             id_=x['ID'],
                                                             fieldname=x.get('fieldname'),
                                                             **{self.secondary_lookup_field: secondary_lookup_value})
                if secondary_lookup_value in key_map:
                    non_unique.add(secondary_lookup_value)
                key_map[secondary_lookup_value] = new_object
            else:
                new_object = self.xnat_session.create_object(x['URI'],
                                                             type_=xsi_type,
                                                             id_=x['ID'],
                                                             fieldname=x.get('fieldname'))

            listing.append(new_object)
            id_map[x['ID']] = new_object

        return id_map, key_map, non_unique, listing

    def tabulate(self, columns=None, filter=None):
        """
        Create a table (tuple of namedtuples) from this listing. It is possible
        to choose the columns and add a filter to the tabulation.

        :param tuple columns: names of the variables to use for columns
        :param dict filter: update filters to use (form of {'variable': 'filter*'}),
                             setting this option will try to merge the filters and
                             throw an error if that is not possible.
        :return: tabulated data
        :rtype: tuple
        :raises ValueError: if the new filters conflict with the object filters
        """
        if columns is None:
            columns = ('DEFAULT',)

        if filter is None:
            filter = self.used_filters
        else:
            filter = self.merge_filters(self.used_filters, filter)

        query = dict(filter)
        query['columns'] = ','.join(columns)

        result = self.xnat_session.get_json(self.uri, query=query)
        if len(result['ResultSet']['Result']) > 0:
            result_columns = list(result['ResultSet']['Result'][0].keys())

            # Retain requested order
            if columns != ('DEFAULT',):
                result_columns = [x for x in columns if x in result_columns]

            # Replace all non-alphanumeric characters with an underscore
            result_columns = {s: re.sub('[^0-9a-zA-Z]+', '_', s) for s in result_columns}
            rowtype = namedtuple('TableRow', list(result_columns.values()))

            # Replace all non-alphanumeric characters in each key of the keyword dictionary
            return tuple(rowtype(**{result_columns[k]: v for k, v in x.items() if k in result_columns}) for x in result['ResultSet']['Result'])
        else:
            return ()

    @property
    def used_filters(self):
        return self._used_filters

    @staticmethod
    def merge_filters(old_filters, extra_filters):
        # First check for conflicting filters
        for key in extra_filters:
            if key in old_filters and old_filters[key] != extra_filters[key]:
                raise ValueError('Trying to redefine filter {key}={oldval} to {key}={newval}'.format(key=key,
                                                                                                     oldval=old_filters[key],
                                                                                                     newval=extra_filters[key]))

        new_filters = dict(old_filters)
        new_filters.update(extra_filters)

        return new_filters

    def filter(self, filters=None, **kwargs):
        """
        Create a new filtered listing based on this listing. There are two way
        of defining the new filters. Either by passing a dict as the first
        argument, or by adding filters as keyword arguments.

        For example::
          >>> listing.filter({'ID': 'A*'})
          >>> listing.filter(ID='A*')

        are equivalent.

        :param dict filters: a dictionary containing the filters
        :param str kwargs: keyword arguments containing the filters
        :return: new filtered XNATListing
        :rtype: XNATListing
        """
        if filters is None:
            filters = kwargs

        new_filters = self.merge_filters(self.used_filters, filters)
        return XNATListing(uri=self.uri,
                           xnat_session=self.xnat_session,
                           parent=self.parent,
                           field_name=self.field_name,
                           secondary_lookup_field=self.secondary_lookup_field,
                           xsi_type=self._xsi_type,
                           filter=new_filters)


class XNATSimpleListing(XNATBaseListing, MutableMapping, MutableSequence):
    def __str__(self):
        if self.secondary_lookup_field is not None:
            content = ', '.join('{!r}: {!r}'.format(key, value) for key, value in self.items())
            content = '{{{}}}'.format(content)
        else:
            content = ', '.join(repr(v) for v in self.values())
            content = '[{}]'.format(content)
        return '<{} {}>'.format(type(self).__name__, content)

    def __iter__(self):
        for key in self.key_map:
            yield key

    @property
    def xnat_session(self):
        return self.parent.xnat_session

    @property
    def fulldata(self):
        for child in self.parent.fulldata['children']:
            if child['field'] == self.field_name:
                return child['items']
        return []

    @property
    @caching
    def data_maps(self):
        id_map = {}
        key_map = {}
        listing = []
        non_unique_keys = set()

        for index, element in enumerate(self.fulldata):
            if self.secondary_lookup_field is not None:
                key = element['data_fields'][self.secondary_lookup_field]
            else:
                key = index

            try:
                value = element['data_fields'][self.field_name.split('/')[-1]]
            except KeyError:
                continue

            if key in key_map:
                non_unique_keys.add(key)
                key_map[key] = None
            elif self.secondary_lookup_field is not None:
                key_map[key] = value

            listing.append(value)

        return id_map, key_map, non_unique_keys, listing

    def __setitem__(self, key, value):
        query = {'xsiType': self.parent.__xsi_type__,
                 '{type_}/{fieldname}[{lookup}={key}]/{fieldpart}'.format(type_=self.parent.__xsi_type__,
                                                                          fieldname=self.field_name,
                                                                          lookup=self.secondary_lookup_field,
                                                                          fieldpart=self.field_name.split('/')[-1],
                                                                          key=key): value}
        self.xnat_session.put(self.parent.fulluri, query=query)

        # Remove cache and make sure the reload the data
        self.clearcache()

    def __delitem__(self, key):
        query = {
            'xsiType': self.parent.__xsi_type__,
            '{type_}/{fieldname}[{lookup}={key}]/{fieldpart}'.format(type_=self.parent.__xsi_type__,
                                                                     fieldname=self.field_name,
                                                                     lookup=self.secondary_lookup_field,
                                                                     fieldpart=self.field_name.split('/')[-1],
                                                                     key=key): 'NULL',
            '{type_}/{fieldname}[{lookup}={key}]/{lookup}'.format(type_=self.parent.__xsi_type__,
                                                                  fieldname=self.field_name,
                                                                  lookup=self.secondary_lookup_field,
                                                                  key=key): 'NULL',
        }
        self.xnat_session.put(self.parent.fulluri, query=query)

        # Remove cache and make sure the reload the data
        self.clearcache()

    def insert(self, index, value):
        pass


class XNATSubListing(XNATBaseListing, MutableMapping, MutableSequence):
    def __getitem__(self, item):
        try:
            return super(XNATSubListing, self).__getitem__(item)
        except (IndexError, KeyError):
            cls = self.xnat_session.XNAT_CLASS_LOOKUP[self._xsi_type]
            object = cls(uri=self.parent.uri, id_=item, datafields={}, parent=self, fieldname=item)
            return object

    @property
    def xnat_session(self):
        return self.parent.xnat_session

    @property
    def fulldata(self):
        for child in self.parent.fulldata['children']:
            if child['field'] == self.field_name or child['field'].startswith(self.field_name + '/'):
                return child['items']
        return []

    @property
    def uri(self):
        return self.parent.fulluri

    @property
    def fulluri(self):
        return self.parent.fulluri

    @property
    @caching
    def data_maps(self):
        id_map = {}
        key_map = {}
        listing = []
        non_unique_keys = set()

        for index, element in enumerate(self.fulldata):
            if self.secondary_lookup_field is not None:
                key = element['data_fields'][self.secondary_lookup_field]
            else:
                key = index

            try:
                xsi_type = element['meta']['xsi:type']
            except KeyError:
                xsi_type = self._xsi_type

            # XNAT seems to like to sometimes give a non-defined XSI type back
            #  (e.g. 'xnat:fieldDefinitionGroup_field'), make sure the XNAT
            # reply contains a valid XSI
            if xsi_type not in self.xnat_session.XNAT_CLASS_LOOKUP:
                xsi_type = self._xsi_type

            cls = self.xnat_session.XNAT_CLASS_LOOKUP[xsi_type]
            object = cls(uri=self.parent.uri,
                         id_=key,
                         datafields=element['data_fields'],
                         parent=self,
                         fieldname=key)

            if key in key_map:
                non_unique_keys.add(key)
                key_map[key] = None
            elif self.secondary_lookup_field is not None:
                key_map[key] = object

            listing.append(object)

        return id_map, key_map, non_unique_keys, listing

    @property
    def __xsi_type__(self):
        return self._xsi_type

    @property
    def xpath(self):
        return '{}/{}'.format(self.parent.xpath, self.field_name)

    def __setitem__(self, key, value):
        pass

    def __delitem__(self, key):
        # Determine XPATH of item to remove
        if isinstance(key, int):
            xpath = '{}[{}]'.format(self.xpath,
                                   key)
        else:
            xpath = '{}[{}={}]'.format(self.xpath,
                                      self.secondary_lookup_field,
                                      key)

        # Get correct xsi type
        if self.parent is not None:
            xsi_type = self.parent.__xsi_type__
        else:
            xsi_type = self.__xsi_type__

        query = {
            'xsiType': xsi_type,
            xpath: 'NULL'
        }

        self.xnat_session.put(self.fulluri, query=query)
        self.clearcache()

    def insert(self, index, value):
        pass



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import datetime
import mimetypes
import collections
import os
import tempfile
from zipfile import ZipFile, ZIP_DEFLATED

import six
from six import BytesIO

from .core import XNATBaseObject
from .prearchive import PrearchiveSession
from .exceptions import XNATResponseError, XNATValueError
from .mixin import ProjectData, SubjectData, ExperimentData

TokenResult = collections.namedtuple('TokenResult', ('alias', 'secret'))


class DicomBoxImportRequest(object):
    def __init__(self, uri, xnat_session):
        self._xnat_session = xnat_session
        self._data = None
        self.uri = uri
        self._update_data()

    def _update_data(self):
        self._data = self._xnat_session.get_json(self.uri)

    @property
    def status(self):
        self._update_data()
        return self._data['status']

    @property
    def id(self):
        return self._data['id']

    @property
    def username(self):
        return self._data['username']

    @property
    def project_id(self):
        return self._data['parameters'].get('PROJECT_ID')

    @property
    def subject_id(self):
        return self._data['parameters'].get('SUBJECT_ID')

    @property
    def session_path(self):
        return self._data['sessionPath']

    @property
    def cleanup_after_import(self):
        return self._data['clenaupAfterImport']

    @property
    def enabled(self):
        return self._data['enabled']

    @property
    def created(self):
        return datetime.datetime.fromtimestamp(self._data['created'] / 1000.0)

    @property
    def timestamp(self):
        return datetime.datetime.fromtimestamp(self._data['timestamp'] / 1000.0)


class Services(object):
    """
    The class representing all service functions in XNAT found in the
    /data/services REST directory
    """
    def __init__(self, xnat_session):
        self._xnat_session = xnat_session

    @property
    def xnat_session(self):
        return self._xnat_session

    def guess_content_type(self, path):
        content_type, _ = mimetypes.guess_type(path)

        if content_type not in ['application/x-tar', 'application/zip']:
            if path.endswith('.zip'):
                self.xnat_session.logger.warning(
                    'Found unexpect content type, but assuming zip based on the extension'
                )
                content_type = 'application/zip'
            elif path.endswith('.tar.gz'):
                self.xnat_session.logger.warning(
                    'Found unexpect content type, but assuming tar based on the extension'
                )
                content_type = 'application/x-tar'
            else:
                self.xnat_session.logger.warning(
                    'Found unexpected content (found type {}), this could result in errors on import!'.format(content_type)
                )

        return content_type

    def dicom_dump(self, src, fields=None):
        """
        Retrieve a dicom dump as a JSON data structure
        See the XAPI documentation for more detailed information: `DICOM Dump Service <https://wiki.xnat.org/display/XAPI/DICOM+Dump+Service+API>`_

        :param str src: The url of the scan to generate the DICOM dump for
        :param list fields: Fields to filter for DICOM tags. It can either a tag name or tag number in the format GGGGEEEE (G = Group number, E = Element number)
        :return: JSON object (dict) representation of DICOM header
        :rtype: dict
        """
        query_string = {'src': src}
        if fields is not None:
            if not isinstance(fields, (list, str)):
                raise XNATValueError('The fields argument to .dicom_dump() should be list or a str and not {}'.format(type(fields)))
            query_string['field'] = fields

        return self.xnat_session.get_json('/data/services/dicomdump', query=query_string)['ResultSet']['Result']

    def import_(self, path, overwrite=None, quarantine=False, destination=None,
                trigger_pipelines=None, project=None, subject=None,
                experiment=None, content_type=None, import_handler=None):
        """
        Import a file into XNAT using the import service. See the
        `XNAT wiki <https://wiki.xnat.org/pages/viewpage.action?pageId=6226268>`_
        for a detailed explanation.

        :param str path: local path of the file to upload and import
        :param str overwrite: how the handle existing data (none, append, delete)
        :param bool quarantine: flag to indicate session should be quarantined
        :param bool trigger_pipelines: indicate that archiving should trigger pipelines
        :param str destination: the destination to upload the scan to
        :param str project: the project in the archive to assign the session to
                            (only accepts project ID, not a label)
        :param str subject: the subject in the archive to assign the session to
        :param str experiment: the experiment in the archive to assign the session content to
        :param str content_type: overwite the content_type (by default the mimetype will be
                                 guessed using the ``mimetypes`` package).
                                 This will often be ``application/zip``.
        :return:

        .. note::
            The project has to be given using the project ID and *NOT* the label.

        .. warning::
            On some systems the guessed mimetype of a zip file might not be ``application/zip``
            but be something like ``application/x-zip-compressed``. In that case you might have to
            set the ``content_type`` parameter to ``application/zip`` manually.

        """
        query = {}
        if overwrite is not None:
            if overwrite not in ['none', 'append', 'delete']:
                raise ValueError('Overwrite should be none, append or delete!')
            query['overwrite'] = overwrite

        if quarantine:
            query['quarantine'] = 'true'

        if trigger_pipelines is not None:
            if isinstance(trigger_pipelines, bool):
                if trigger_pipelines:
                    query['triggerPipelines'] = 'true'
                else:
                    query['triggerPipelines'] = 'false'
            else:
                raise TypeError('trigger_pipelines should be a boolean')

        if destination is not None:
            query['dest'] = destination

        if project is not None:
            if isinstance(project, ProjectData):
                project = project.id
            query['project'] = project

        if subject is not None:
            if isinstance(subject, SubjectData):
                subject = subject.label
            query['subject'] = subject

        if experiment is not None:
            if isinstance(experiment, ExperimentData):
                experiment = experiment.label
            query['session'] = experiment

        if import_handler is not None:
            query['import-handler'] = import_handler

        # Get mimetype of file
        if content_type is None and isinstance(path, six.string_types):
            content_type = self.guess_content_type(path)

        uri = '/data/services/import'
        response = self.xnat_session.upload(uri=uri, file_=path, query=query, content_type=content_type, method='post')

        if response.status_code != 200:
            raise XNATResponseError('The response for uploading was ({}) {}'.format(response.status_code, response.text))

        # Create object, the return text should be the url, but it will have a \r\n at the end that needs to be stripped
        response_text = response.text.strip()
        if response_text.startswith('/data/prearchive'):
            return PrearchiveSession(response_text, self.xnat_session)

        return self.xnat_session.create_object(response_text)

    def _zip_directory(self, directory, fh):
        """
        Zip a directory into a file(-like) obj given

        :param directory: directory to zip
        :param fh: output file handle
        """
        with ZipFile(fh, 'w', ZIP_DEFLATED) as zip_file:
            for dirpath, dirs, files in os.walk(directory):
                for f in files:
                    zip_file.write(os.path.join(dirpath, f),
                                   os.path.relpath(os.path.join(dirpath, f),
                                                   os.path.dirname(directory)))

    def import_dir(self, directory, overwrite=None, quarantine=False, destination=None,
                trigger_pipelines=None, project=None, subject=None,
                experiment=None, method='zip_file', import_handler=None):
        """
        Import a directory to an XNAT resource.

        :param str directory: local path of the directory to upload and import
        :param str overwrite: how the handle existing data (none, append, delete)
        :param bool quarantine: flag to indicate session should be quarantined
        :param bool trigger_pipelines: indicate that archiving should trigger pipelines
        :param str destination: the destination to upload the scan to
        :param str project: the project in the archive to assign the session to
                            (only accepts project ID, not a label)
        :param str subject: the subject in the archive to assign the session to
        :param str experiment: the experiment in the archive to assign the session content to


        The method has 2 options, default is zip_file:

        #. ``zip_file``: Create a temporary zip file and upload that
        #. ``zip_memory``: Create a temporary zip file in memory and upload it

        The considerations are that sometimes you can fit things in memory so
        you can save disk IO by putting it in memory.

        """
        # Make sure that a None or empty string is replaced by the default
        method = method or 'zip_file'

        content_type = 'application/zip'
        if method == 'zip_file':
            # Max-size is 256 MB
            fh = tempfile.SpooledTemporaryFile(max_size=268435456, mode='wb+')
            self._zip_directory(directory=directory, fh=fh)
        elif method == 'zip_memory':
            fh = BytesIO()
            self._zip_directory(directory=directory, fh=fh)
        else:
            message = 'Selected invalid import directory method: {}!'.format(method)
            self.xnat_session.logger.error(message)
            raise XNATValueError(message)

        fh.seek(0)
        session = self.import_(fh, overwrite, quarantine, destination,
                               trigger_pipelines, project, subject,
                               experiment, content_type, import_handler=import_handler)
        fh.close()
        return session

    def import_dicom_inbox(self, path, cleanup=False, project=None, subject=None, experiment=None):
        """
        Import a file into XNAT using the import service. See the
        `XNAT wiki <https://wiki.xnat.org/pages/viewpage.action?pageId=6226268>`_
        for a detailed explanation.

        :param str path: local path of the file to upload and import
        :param str cleanup: remove the files after importing them (default false)
        :param str project: the project in the archive to assign the session to
                            (only accepts project ID, not a label)
        :param str subject: the subject in the archive to assign the session to
        :param str experiment: the experiment in the archive to assign the session content to
        :return:

        .. note::
            The project and subject has to be given using the ID and *NOT* the label/name.
        """
        query = {
            'import-handler': 'inbox',
            'path': path,
        }

        if cleanup:
            query['cleanupAfterImport'] = 'true'
        else:
            query['cleanupAfterImport'] = 'false'

        if project is not None:
            if isinstance(project, ProjectData):
                project = project.id
            query['PROJECT_ID'] = project

        if subject is not None:
            if isinstance(subject, SubjectData):
                subject = subject.id
            query['SUBJECT_ID'] = subject

        if experiment is not None:
            if isinstance(experiment, ExperimentData):
                experiment = experiment.label
            query['EXPT_LABEL'] = experiment

        uri = '/data/services/import'
        response = self.xnat_session.post(uri, query=query)

        if response.status_code != 200:
            raise XNATResponseError('The response for uploading was ({}) {}'.format(response.status_code, response.text))

        # Create object, the return text should be the url, but it will have a \r\n at the end that needs to be stripped
        response_text = response.text.strip()
        if response_text.startswith('/data/prearchive'):
            return PrearchiveSession(response_text, self.xnat_session)

        return DicomBoxImportRequest(response_text, self._xnat_session)

    def issue_token(self, user=None):
        """
        Issue a login token, by default for the current logged in user. If
        username is given, for that user. To issue tokens for other users
        you must be an admin.

        :param str user: User to issue token for, default is current user
        :return: Token in a named tuple (alias, secret)
        """
        uri = '/data/services/tokens/issue'
        if user:
            uri += '/user/{}'.format(user)

        result = self.xnat_session.get_json(uri)

        return TokenResult(result['alias'], result['secret'])

    def refresh_catalog(self, resource, checksum=False, delete=False, append=False, populate_stats=False):
        """
        Call for a refresh of the catalog, see https://wiki.xnat.org/display/XAPI/Catalog+Refresh+API for
        details.

        Introduced with XNAT 1.6.2, the refresh catalog service is used to update catalog xmls that are out of
        sync with the file system.  This service can be used to store checksums for entries that are missing the,
        remove entries that no longer have valid files, or add new entries for files that have been manually
        added to the archive directory.

        When using this feature to add files that have been manually added to the archive directory, you must
        have placed the files in the appropriate archive directory (in the same directory as the generated
        catalog xml or a sub-directory).  The catalog xml should already exist before triggering this service.
        If you haven't generated the catalog yet, you can do so by doing a PUT to the resource URL
        (i.e. /data/archive/experiments/ID/resources/TEST).

        Extra parameters indicate operations to perform on the specified resource(s) during the refresh.
        If non are given, then the catalog will be reviewed and updated for validity, but nothing else.

        :param resource: XNATObject or uri indicating the resource to use
        :param bool checksum: generate checksums for any entries that are missing them
        :param bool delete: remove entries that do not reference valid files
        :param bool append: add entries for files in the catalog directory (or sub-directory)
        :param bool populate_stats: updates the statistics for the resource in the XNAT abstract resource table.
        :return:
        """
        if isinstance(resource, XNATBaseObject):
            resource = resource.fulluri

        options = []

        # Check which options to add
        if checksum:
            options.append('checksum')

        if delete:
            options.append('delete')

        if append:
            options.append('append')

        if populate_stats:
            options.append('populateStats')

        # Make query that only contains options if any are given
        query = {'resource': resource}

        if options:
            query['options'] = options

        self.xnat_session.post('/data/services/refresh/catalog', query=query)



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import os
import tempfile
from zipfile import ZipFile
import tarfile
import shutil

from six import BytesIO

from .core import caching, XNATBaseObject, XNATListing
from .search import SearchField
from .utils import mixedproperty

try:
    PYDICOM_LOADED = True
    import pydicom
except ImportError:
    PYDICOM_LOADED = False


class ProjectData(XNATBaseObject):
    SECONDARY_LOOKUP_FIELD = 'name'

    @property
    def fulluri(self):
        return '{}/projects/{}'.format(self.xnat_session.fulluri, self.id)

    @property
    @caching
    def subjects(self):
        return XNATListing(self.uri + '/subjects',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='subjects',
                           secondary_lookup_field='label',
                           xsi_type='xnat:subjectData')

    @property
    @caching
    def experiments(self):
        return XNATListing(self.uri + '/experiments',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='experiments',
                           secondary_lookup_field='label')

    @property
    @caching
    def files(self):
        return XNATListing(self.uri + '/files',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='files',
                           secondary_lookup_field='path',
                           xsi_type='xnat:fileData')

    @property
    @caching
    def resources(self):
        return XNATListing(self.uri + '/resources',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='resources',
                           secondary_lookup_field='label',
                           xsi_type='xnat:resourceCatalog')

    def create_resource(self, label, format=None, data_dir=None, method=None):
        uri = '{}/resources/{}'.format(self.fulluri, label)
        self.xnat_session.put(uri, format=format)
        self.clearcache()  # The resources changed, so we have to clear the cache
        resource = self.xnat_session.create_object(uri, type_='xnat:resourceCatalog')

        if data_dir is not None:
            resource.upload_dir(data_dir, method=method)

        return resource

    def download_dir(self, target_dir, verbose=True, progress_callback=None):
        """
        Download the entire project and unpack it in a given directory. Note
        that this method will create a directory structure following
        $target_dir/{project.name}/{subject.label}/{experiment.label}
        and unzip the experiment zips as given by XNAT into that. If
        the $target_dir/{project.name} does not exist, it will be created.

        :param str target_dir: directory to create project directory in
        :param bool verbose: show progress
        :param progress_callback: function to call with progress string
                                  should be a function with one argument
        """

        project_dir = os.path.join(target_dir, self.name)
        if not os.path.isdir(project_dir):
            os.mkdir(project_dir)

        number_of_subjects = len(self.subjects)

        for n, subject in enumerate(self.subjects.values(), start=1):
            if progress_callback:
                progress_callback("Downloading subject {} of {}".format(n, number_of_subjects))

            subject.download_dir(project_dir, verbose=verbose, progress_callback=progress_callback)

        if verbose:
            self.logger.info('Downloaded project to {}'.format(project_dir))


class InvestigatorData(XNATBaseObject):
    def __str__(self):
        title = self.title or ''
        first = self.firstname or ''
        last = self.lastname or ''
        fullname = '{title} {first} {last}'.format(title=title, first=first, last=last).strip()
        return '<InvestigatorData {fullname}>'.format(fullname=fullname)


class SubjectData(XNATBaseObject):
    SECONDARY_LOOKUP_FIELD = 'label'

    @property
    def fulluri(self):
        return '{}/projects/{}/subjects/{}'.format(self.xnat_session.fulluri, self.project, self.id)

    @mixedproperty
    def label(cls):
        # 0 Automatically generated Property, type: xs:string
        return SearchField(cls, "label")

    @label.getter
    def label(self):
        # Check if label is already inserted during listing, that should be valid
        # label for the project under which it was listed in the first place
        try:
            return self._overwrites['label']
        except KeyError:
            pass

        # Retrieve the label the hard and costly way
        try:
            # First check if subject is shared into current project
            sharing = next(x for x in self.fulldata['children'] if x['field'] == 'sharing/share')
            share_info = next(x for x in sharing['items'] if x['data_fields']['project'] == self.project)
            label = share_info['data_fields']['label']
        except (KeyError, StopIteration):
            label = self.get('label', type_=str)

        # Cache label for future use
        self._overwrites['label'] = label
        return label

    @label.setter
    def label(self, value):
        self.xnat_session.put(self.fulluri, query={'label': value})
        self.clearcache()

    @property
    @caching
    def files(self):
        return XNATListing(self.uri + '/files',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='files',
                           secondary_lookup_field='path',
                           xsi_type='xnat:fileData')

    def download_dir(self, target_dir, verbose=True, progress_callback=None):
        """
        Download the entire subject and unpack it in a given directory. Note
        that this method will create a directory structure following
        $target_dir/{subject.label}/{experiment.label}
        and unzip the experiment zips as given by XNAT into that. If
        the $target_dir/{subject.label} does not exist, it will be created.

        :param str target_dir: directory to create subject directory in
        :param bool verbose: show progress
        :param progress_callback: function to call with progress string
                                  should be a function with one argument
        """
        subject_dir = os.path.join(target_dir, self.label)
        if not os.path.isdir(subject_dir):
            os.mkdir(subject_dir)

        number_of_experiments = len(self.experiments)

        for n, experiment in enumerate(self.experiments.values(), start=1):
            if progress_callback is not None:
                progress_callback("Downloading experiment {} of {}".format(n, number_of_experiments))
            experiment.download_dir(subject_dir, verbose=verbose)

        if verbose:
            self.logger.info('Downloaded subject to {}'.format(subject_dir))

    def share(self, project, label=None):
        # Create the uri for sharing
        share_uri = '{}/projects/{}'.format(self.fulluri, project)

        # Add label if needed
        query = {}
        if label is not None:
            query['label'] = label

        self.xnat_session.put(share_uri, query=query)
        self.clearcache()

    def create_resource(self, label, format=None, data_dir=None, method=None):
        uri = '{}/resources/{}'.format(self.fulluri, label)
        self.xnat_session.put(uri, format=format)
        self.clearcache()  # The resources changed, so we have to clear the cache
        resource = self.xnat_session.create_object(uri, type_='xnat:resourceCatalog')

        if data_dir is not None:
            resource.upload_dir(data_dir, method=method)

        return resource


class ExperimentData(XNATBaseObject):
    SECONDARY_LOOKUP_FIELD = 'label'

    @mixedproperty
    def label(cls):
        return SearchField(cls, "label")

    @label.getter
    def label(self):
        # Check if label is already inserted during listing, that should be valid
        # label for the project under which it was listed in the first place
        try:
            return self._overwrites['label']
        except KeyError:
            pass

        # Retrieve the label the hard and costly way
        try:
            # First check if subject is shared into current project
            sharing = next(x for x in self.fulldata['children'] if x['field'] == 'sharing/share')
            share_info = next(x for x in sharing['items'] if x['data_fields']['project'] == self.project)
            label = share_info['data_fields']['label']
        except (KeyError, StopIteration):
            label = self.get('label', type_=str)

        # Cache label for future use
        self._overwrites['label'] = label
        return label

    @label.setter
    def label(self, value):
        self.xnat_session.put(self.fulluri, query={'label': value})
        self.clearcache()


class SubjectAssessorData(XNATBaseObject):
    @property
    def fulluri(self):
        return '/data/archive/projects/{}/subjects/{}/experiments/{}'.format(self.project, self.subject_id, self.id)

    @property
    def subject(self):
        return self.xnat_session.subjects[self.subject_id]


class ImageSessionData(XNATBaseObject):
    @property
    @caching
    def files(self):
        return XNATListing(self.fulluri + '/files',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='files',
                           secondary_lookup_field='path',
                           xsi_type='xnat:fileData')

    def create_assessor(self, label, type_):
        uri = '{}/assessors/{label}?xsiType={type}&label={label}&req_format=qs'.format(self.fulluri,
                                                                                       type=type_,
                                                                                       label=label)
        self.xnat_session.put(uri, accepted_status=(200, 201))
        self.clearcache()  # The resources changed, so we have to clear the cache
        return self.xnat_session.create_object('{}/assessors/{}'.format(self.fulluri, label), type_=type_)

    def download(self, path, verbose=True):
        self.xnat_session.download_zip(self.fulluri + '/scans/ALL/files', path, verbose=verbose)

    def download_dir(self, target_dir, verbose=True):
        """
        Download the entire experiment and unpack it in a given directory. Note
        that this method will create a directory structure following
        $target_dir/{experiment.label} and unzip the experiment zips
        as given by XNAT into that. If the $target_dir/{experiment.label} does
        not exist, it will be created.

        :param str target_dir: directory to create experiment directory in
        :param bool verbose: show progress
        """
        # Check if there are actually file to be found
        file_list = self.xnat_session.get_json(self.fulluri + '/scans/ALL/files')
        if len(file_list['ResultSet']['Result']) == 0:
            # Just make sure the target directory exists and stop
            if not os.path.exists(target_dir):
                os.mkdir(target_dir)
            return

        with tempfile.TemporaryFile() as temp_path:
            self.xnat_session.download_stream(self.fulluri + '/scans/ALL/files', temp_path, format='zip', verbose=verbose)

            with ZipFile(temp_path) as zip_file:
                zip_file.extractall(target_dir)

        if verbose:
            self.logger.info('\nDownloaded image session to {}'.format(target_dir))

    def share(self, project, label=None):
        # Create the uri for sharing
        share_uri = '{}/projects/{}'.format(self.fulluri, project)

        # Add label if needed
        query = {}
        if label is not None:
            query['label'] = label

        self.xnat_session.put(share_uri, query=query)
        self.clearcache()


class DerivedData(XNATBaseObject):
    @property
    def fulluri(self):
        return '/data/experiments/{}/assessors/{}'.format(self.image_session_id, self.id)

    @property
    @caching
    def files(self):
        return XNATListing(self.fulluri + '/files',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='files',
                           secondary_lookup_field='path',
                           xsi_type='xnat:fileData')

    @property
    @caching
    def resources(self):
        return XNATListing(self.fulluri + '/resources',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='resources',
                           secondary_lookup_field='label',
                           xsi_type='xnat:resourceCatalog')

    def create_resource(self, label, format=None, data_dir=None, method=None):
        uri = '{}/resources/{}'.format(self.fulluri, label)
        self.xnat_session.put(uri, format=format)
        self.clearcache()  # The resources changed, so we have to clear the cache
        resource = self.xnat_session.create_object(uri, type_='xnat:resourceCatalog')

        if data_dir is not None:
            resource.upload_dir(data_dir, method=method)

        return resource

    def download(self, path, verbose=True):
        self.xnat_session.download_zip(self.uri + '/files', path, verbose=verbose)


class ImageScanData(XNATBaseObject):
    SECONDARY_LOOKUP_FIELD = 'type'

    @property
    @caching
    def files(self):
        return XNATListing(self.uri + '/files',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='files',
                           secondary_lookup_field='path',
                           xsi_type='xnat:fileData')

    @property
    @caching
    def resources(self):
        return XNATListing(self.uri + '/resources',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='resources',
                           secondary_lookup_field='label',
                           xsi_type='xnat:resourceCatalog')

    def create_resource(self, label, format=None, data_dir=None, method='tgz_file'):
        uri = '{}/resources/{}'.format(self.uri, label)
        self.xnat_session.put(uri, format=format)
        self.clearcache()  # The resources changed, so we have to clear the cache
        resource = self.xnat_session.create_object(uri, type_='xnat:resourceCatalog')

        if data_dir is not None:
            resource.upload_dir(data_dir, method=method)

        return resource

    def download(self, path, verbose=True):
        self.xnat_session.download_zip(self.uri + '/files', path, verbose=verbose)

    def download_dir(self, target_dir, verbose=True):
        with tempfile.TemporaryFile() as temp_path:
            self.xnat_session.download_stream(self.uri + '/files', temp_path, format='zip', verbose=verbose)

            with ZipFile(temp_path) as zip_file:
                zip_file.extractall(target_dir)

        if verbose:
            self.logger.info('Downloaded image scan data to {}'.format(target_dir))

    def dicom_dump(self, fields=None):
        """
        Retrieve a dicom dump as a JSON data structure
        See the XAPI documentation for more detailed information: `DICOM Dump Service <https://wiki.xnat.org/display/XAPI/DICOM+Dump+Service+API>`_

        :param list fields: Fields to filter for DICOM tags. It can either a tag name or tag number in the format GGGGEEEE (G = Group number, E = Element number)
        :return: JSON object (dict) representation of DICOM header
        :rtype: dict
        """
        experiment = self.xnat_session.create_object('/data/experiments/{}'.format(self.image_session_id))

        uri = '/archive/projects/{}/experiments/{}/scans/{}'.format(
            experiment.project,
            self.image_session_id,
            self.id,
        )
        return self.xnat_session.services.dicom_dump(src=uri, fields=fields)

    def read_dicom(self, file=None, read_pixel_data=False, force=False):
        # Check https://gist.github.com/obskyr/b9d4b4223e7eaf4eedcd9defabb34f13 for partial loading?
        if not PYDICOM_LOADED:
            raise RuntimeError('Cannot read DICOM, missing required dependency: pydicom')

        dicom_resource = self.resources.get('DICOM',
                                            self.resources.get('secondary'))

        if dicom_resource is None:
            raise ValueError('Scan {} does not contain a DICOM resource!'.format(self))

        if file is None:
            dicom_files = sorted(dicom_resource.files.values(), key=lambda x: x.path)
            file = dicom_files[0]
        else:
            if file not in dicom_resource.files.values():
                raise ValueError('File {} not part of scan {} DICOM resource'.format(file, self))

        with file.open() as dicom_fh:
            dicom_data = pydicom.dcmread(dicom_fh,
                                         stop_before_pixels=not read_pixel_data,
                                         force=force)

        return dicom_data


class AbstractResource(XNATBaseObject):
    SECONDARY_LOOKUP_FIELD = 'label'

    def __init__(self,
                 uri=None,
                 xnat_session=None,
                 id_=None,
                 datafields=None,
                 parent=None,
                 fieldname=None,
                 overwrites=None,
                 data_dir=None,
                 upload_method=None,
                 **kwargs):

        super(AbstractResource, self).__init__(
            uri=uri,
            xnat_session=xnat_session,
            id_=id_,
            datafields=datafields,
            parent=parent,
            fieldname=fieldname,
            overwrites=overwrites,
            **kwargs
        )

        if data_dir is not None:
            self.upload_dir(data_dir, method=upload_method)

    @property
    @caching
    def fulldata(self):
        # FIXME: ugly hack because direct query fails
        uri, label = self.uri.rsplit('/', 1)
        data = self.xnat_session.get_json(uri)['ResultSet']['Result']

        def _guess_key( d ):
            if 'URI' not in d and 'ID' not in d and 'xnat_abstractresource_id' in d:
                # HACK: This is a Resource where the label is not part of the uri, it uses this xnat_abstractresource_id instead.
                return d['xnat_abstractresource_id']
            else:
                return d['label']

        try:
            data = next(x for x in data if _guess_key(x) == label)
        except StopIteration:
            raise ValueError('Cannot find full data!')

        data['ID'] = data['xnat_abstractresource_id']  # Make sure the ID is present
        return data

    @property
    def data(self):
        return self.fulldata

    @property
    def file_size(self):
        file_size = self.data['file_size']
        if file_size.strip() == '':
            return 0
        else:
            return int(file_size)

    @property
    def file_count(self):
        file_count = self.data['file_count']
        if file_count.strip() == '':
            return 0
        else:
            return int(file_count)

    def refresh_catalog(self):
        """
        Call refresh catalog on this object, see :py:meth:`xnat.services.Services.refresh_catalog` for details.
        """
        self.xnat_session.services.refresh_catalog(self)

    @property
    @caching
    def files(self):
        return XNATListing(self.uri + '/files',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='files',
                           secondary_lookup_field='path',
                           xsi_type='xnat:fileData')

    def download(self, path, verbose=True):
        self.xnat_session.download_zip(self.uri + '/files', path, verbose=verbose)

    def download_dir(self, target_dir, verbose=True, flatten_dirs=False):
        """
        Download the entire resource and unpack it in a given directory

        :param str target_dir: directory to unpack to
        :param bool verbose: show progress
        """
        with tempfile.TemporaryFile() as temp_path:
            self.xnat_session.download_stream(self.uri + '/files', temp_path, format='zip', verbose=verbose)

            with ZipFile(temp_path) as zip_file:
                zip_file.extractall(target_dir)
                extracted_files = zip_file.namelist()

            extraction_sub_directories = [os.path.dirname(os.path.normpath(i_extracted_file)) for i_extracted_file in extracted_files]
            unique_extraction_sub_directories = list(set(extraction_sub_directories))
            # Check if we have multiple resources
            multiple_resources = len(unique_extraction_sub_directories) > 1

            if flatten_dirs:
                for i_extracted_file in extracted_files:
                    new_resource_path = target_dir
                    if multiple_resources:
                        # With multiple resources we keep the subfolder
                        new_resource_path = os.path.join(new_resource_path, os.path.basename(os.path.dirname(os.path.normpath(i_extracted_file))))
                    if not os.path.exists(new_resource_path):
                        os.makedirs(new_resource_path)

                    new_resource_path = os.path.join(new_resource_path,  os.path.basename(os.path.normpath(i_extracted_file)))

                    shutil.move(os.path.join(target_dir, i_extracted_file), new_resource_path)

                # Remove the original download directory
                root_extraction_sub_dir = os.path.join(target_dir, os.path.normpath(extraction_sub_directories[0]).split(os.sep)[0])
                shutil.rmtree(root_extraction_sub_dir)
                scan_directory = target_dir
            else:
                if multiple_resources:
                    scan_directory = os.path.join(target_dir, os.path.dirname(os.path.normpath(unique_extraction_sub_directories[0])))
                else:
                    scan_directory = os.path.join(target_dir, unique_extraction_sub_directories[0])

        if verbose:
            self.logger.info('Downloaded resource data to {}'.format(scan_directory))
        return scan_directory

    def upload(self, data, remotepath, overwrite=False, extract=False, file_content=None, file_format=None, file_tags=None, **kwargs):
        """
        Upload a file as an XNAT resource.

        :param str data: The path to the file to upload
        :param str remotepath: The remote path to which to uploadt
        :param bool overwrite: Flag to force overwriting of files
        :param bool extract: Extract the files on the XNAT server
        :param str file_content: Set the Content of the file on XNAT
        :param str file_format: Set the format of the file on XNAT
        :param str file_tags: Set the tags of the file on XNAT
        """
        uri = '{}/files/{}'.format(self.uri, remotepath.lstrip('/'))
        query = {}
        if extract:
            query['extract'] = 'true'
        if file_content is not None:
            query['content'] = file_content
        if file_format is not None:
            query['format'] = file_format
        if file_tags is not None:
            query['tags'] = file_tags

        self.xnat_session.upload(uri, data, overwrite=overwrite, query=query, **kwargs)
        self.files.clearcache()

    def upload_dir(self, directory, overwrite=False, method='tgz_file', **kwargs):
        """
        Upload a directory to an XNAT resource. This means that if you do
        resource.upload_dir(directory) that if there is a file directory/a.txt
        it will be uploaded to resource/files/a.txt

        The method has 5 options, default is tgz_file:

        #. ``per_file``: Scans the directory and uploads file by file
        #. ``tar_memory``: Create a tar archive in memory and upload it in one go
        #. ``tgz_memory``: Create a gzipped tar file in memory and upload that
        #. ``tar_file``: Create a temporary tar file and upload that
        #. ``tgz_file``: Create a temporary gzipped tar file and upload that

        The considerations are that sometimes you can fit things in memory so
        you can save disk IO by putting it in memory. The per file does not
        create additional archives, but has one request per file so might be
        slow when uploading many files.

        :param str directory: The directory to upload
        :param bool overwrite: Flag to force overwriting of files
        :param str method: The method to use
        """
        if not isinstance(directory, str):
            directory = str(directory)

        # Make sure that a None or empty string is replaced by the default
        method = method or 'tgz_file'

        if method == 'per_file':
            for root, _, files in os.walk(directory):
                for filename in files:
                    file_path = os.path.join(root, filename)
                    if os.path.getsize(file_path) == 0:
                        continue

                    target_path = os.path.relpath(file_path, directory)
                    self.upload(file_path, target_path, overwrite=overwrite, **kwargs)
        elif method == 'tar_memory':
            fh = BytesIO()
            with tarfile.open(mode='w', fileobj=fh) as tar_file:
                tar_file.add(directory, '')
            fh.seek(0)
            self.upload(fh, 'upload.tar', overwrite=overwrite, extract=True, **kwargs)
            fh.close()
        elif method == 'tgz_memory':
            fh = BytesIO()
            with tarfile.open(mode='w:gz', fileobj=fh) as tar_file:
                tar_file.add(directory, '')

            fh.seek(0)
            self.upload(fh, 'upload.tar.gz', overwrite=overwrite, extract=True, **kwargs)
            fh.close()
        elif method == 'tar_file':
            # Max-size is 256 MB
            with tempfile.SpooledTemporaryFile(max_size=268435456, mode='wb+') as fh:
                with tarfile.open(mode='w', fileobj=fh) as tar_file:
                    tar_file.add(directory, '')
                fh.seek(0)
                self.upload(fh, 'upload.tar', overwrite=overwrite, extract=True, **kwargs)
        elif method == 'tgz_file':
            # Max-size is 256 MB
            with tempfile.SpooledTemporaryFile(max_size=268435456, mode='wb+') as fh:
                with tarfile.open(mode='w:gz', fileobj=fh) as tar_file:
                        tar_file.add(directory, '')

                fh.seek(0)
                self.upload(fh, 'upload.tar.gz', overwrite=overwrite, extract=True, **kwargs)
        else:
            print('Selected invalid upload directory method!')
